/*-----------------------------------------------------------------------------
  Project:  QRD
  File:     qrd.h

  Copyright (C)2010 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Revision: 1.7 $
 *----------------------------------------------------------------------------*/
#ifndef MW_QRD_QRD_H_
#define MW_QRD_QRD_H_

#include <nn/os.h>

#ifdef __cplusplus

namespace mw {
namespace qrd {

//==============================================================================

const int DATA_MAX = 7089;      //<! QRコード (Version 40) の規格上の格納可能最大データ数

typedef struct tagQRDContext QRDContext;  // モジュールコンテキスト構造体

// QRコードに格納されているデータのタイプ
typedef enum DataType
{
    DATA_TYPE_NUMERIC = 0,      //<! 数字モード
    DATA_TYPE_ALPHABET,         //<! 英数字モード
    DATA_TYPE_KANJI,            //<! 漢字モード (Shift-JISコード)
    DATA_TYPE_BINARY            //<! 8ビットバイナリデータ
} DataType;

// QRコードデコードのエラーコード
typedef enum Result
{
    RESULT_SUCCESS = 0,         //<! 処理成功
    RESULT_ILLEGAL_PARAMETER,   //<! 不正なパラメータ
    RESULT_NODATA,              //<! QRコード認識失敗
    RESULT_NOMEMORY,            //<! 作業用バッファが不足
    RESULT_VERSION,             //<! QRコードのバージョンサポート対象外
    RESULT_ILLEGAL_IMAGESIZE,   //<! イメージのサイズが不正
    RESULT_NOT_INITIALIZED      //<! ライブラリが初期化されていない
} Result;

// QRコードの位置情報を格納する構造体
typedef struct Rect
{
    s32 top;
    s32 left;
    s32 right;
    s32 bottom;
} Rect;

// QRコードをデコードした結果を格納する構造体
typedef struct DecodeData
{
    Rect qrPosition;            //<! QRコードの位置情報
    s32 version;                //<! QRコードのバージョン番号 (1～40)
    s32 size;                   //<! デコードしたデータの有効サイズ
    s32 total;                  //<! 分割QRコードの場合の分割数 (0～15)
    s32 index;                  //<! 分割QRコードの場合のインデックス (0～15)
    s32 parity;                 //<! 分割QRコードの場合のパリティ
    DataType type;              //<! デコードしたQRコードのデータタイプ
    u8 data[DATA_MAX + 1];      //<! デコードしたデータ
    NN_PADDING1;
} DecodeData;

// QRコードデコードを実行する画像を格納する構造体
typedef struct ImageInfo
{
    u16 width;                  //<! デコード対象の画像の幅
    u16 height;                 //<! デコード対象の画像の高さ
    u32* yuv422Data;            //<! デコード対象の画像データ (YUV422フォーマット)
    Rect qrPosition;            //<! デコード対象矩形
} ImageInfo;

//==============================================================================

class QRDecoder
{
public:
    QRDecoder();
    virtual ~QRDecoder();

    bool Initialize(
        void* ptr,
        u32 size);

    void Finalize();

    Result ExecuteDecode(
        DecodeData* result,
        const ImageInfo* img);

    static u32 GetDecodeBufferSize(
        u16 width,
        u16 height);
};

} // namespace qre
} // namespcae mw

#endif // __cplusplus

//==============================================================================

#ifdef __cplusplus
# define MW_QRD_EXTERN_C extern "C"
#else
# define MW_QRD_EXTERN_C extern
#endif

/*!
 * @union mwqrdQRDecoder
 * @brief  QRデコーダをあらわす C の共用体です。
 * @brief  対応するmw::qrd::QRDecoder を参照してください。
 */
typedef union mwqrdQRDecoder
{
	char buf[4];
	u32 alignment_holder;
} mwqrdQRDecoder;

const int MW_QRD_DATA_MAX = 7089;      //<! QRコード (Version 40) の規格上の格納可能最大データ数

// QRコードに格納されているデータのタイプ
typedef enum mwqrdDataType
{
    MW_QRD_DATA_TYPE_NUMERIC = 0,      //<! 数字モード
    MW_QRD_DATA_TYPE_ALPHABET,         //<! 英数字モード
    MW_QRD_DATA_TYPE_KANJI,            //<! 漢字モード (Shift-JISコード)
    MW_QRD_DATA_TYPE_BINARY            //<! 8ビットバイナリデータ
} mwqrdDataType;

// QRコードデコードのエラーコード
typedef enum mwqrdResult
{
    MW_QRD_RESULT_SUCCESS = 0,         //<! 処理成功
    MW_QRD_RESULT_ILLEGAL_PARAMETER,   //<! 不正なパラメータ
    MW_QRD_RESULT_NODATA,              //<! QRコード認識失敗
    MW_QRD_RESULT_NOMEMORY,            //<! 作業用バッファが不足
    MW_QRD_RESULT_VERSION,             //<! QRコードのバージョンサポート対象外
    MW_QRD_RESULT_ILLEGAL_IMAGESIZE,   //<! イメージのサイズが不正
    MW_QRD_RESULT_NOT_INITIALIZED      //<! ライブラリが初期化されていない
} mwqrdResult;

// QRコードの位置情報を格納する構造体
typedef struct mwqrdRect
{
    s32 top;
    s32 left;
    s32 right;
    s32 bottom;
} mwqrdRect;

// QRコードをデコードした結果を格納する構造体
typedef struct mwqrdDecodeData
{
    mwqrdRect qrPosition;            //<! QRコードの位置情報
    s32 version;                     //<! QRコードのバージョン番号 (1～40)
    s32 size;                        //<! デコードしたデータの有効サイズ
    s32 total;                       //<! 分割QRコードの場合の分割数 (0～15)
    s32 index;                       //<! 分割QRコードの場合のインデックス (0～15)
    s32 parity;                      //<! 分割QRコードの場合のパリティ
    mwqrdDataType type;              //<! デコードしたQRコードのデータタイプ
    u8 data[MW_QRD_DATA_MAX + 1];    //<! デコードしたデータ
    NN_PADDING1;
} mwqrdDecodeData;

// QRコードデコードを実行する画像を格納する構造体
typedef struct mwqrdImageInfo
{
    u16 width;                       //<! デコード対象の画像の幅
    u16 height;                      //<! デコード対象の画像の高さ
    u32* yuv422Data;                 //<! デコード対象の画像データ (YUV422フォーマット)
    mwqrdRect qrPosition;            //<! デコード対象矩形
} mwqrdImageInfo;

/*!
 *  @brief 対応する C++ 関数 @ref mw::qrd::QRDecoder::Initialize を参照してください。
 */
MW_QRD_EXTERN_C bool mwqrdQRDecoderInitialize(mwqrdQRDecoder* this_, void* ptr, u32 size);

/*!
 *  @brief 対応する C++ 関数 @ref mw::qrd::QRDecoder::Finalize を参照してください。
 */
MW_QRD_EXTERN_C void mwqrdQRDecoderFinalize(mwqrdQRDecoder* this_);

/*!
 *  @brief 対応する C++ 関数 @ref mw::qrd::QRDecoder::ExecuteDecode を参照してください。
 */
MW_QRD_EXTERN_C mwqrdResult mwqrdQRDecoderExecuteDecode(mwqrdQRDecoder* this_, mwqrdDecodeData* result, const mwqrdImageInfo* img);

/*!
 *  @brief 対応する C++ 関数 @ref mw::qrd::QRDecoder::GetDecodeBufferSize を参照してください。
 */
MW_QRD_EXTERN_C u32 mwqrdQRDecoderGetDecodeBufferSize(u16 width, u16 height);

//==============================================================================

/* MW_QRD_QRD_H_ */
#endif
