﻿/*---------------------------------------------------------------------------*
  Project:  QRD
  File:     main.cpp

  Copyright (C)2010 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/hid.h>
#include <nn/camera.h>
#include <nn/y2r.h>
#include <string.h>

#include <mw/qrd/qrd.h>
#include "GraphicsControl.h"

#define DISP_WIDTH          256    /**< 画面の幅 */
#define DISP_HEIGHT         192    /**< 画面の高さ */
#define CAMERA_WIDTH        640    /**< VGAサイズカメラの幅 */
#define CAMERA_HEIGHT       480    /**< VGAサイズカメラの高さ */
#define IMAGE_WIDTH         320    /**< VGAカメラサイズ時のトリミングイメージ幅 */
#define IMAGE_HEIGHT        240    /**< VGAカメラサイズ時のトリミングイメージ幅 */

#define Y2R_YUV_BYTES       2      /**< YUVバイト数 */
#define Y2R_RGB_BYTES       3      /**< RGBバイト数 */

/** @brief アプリケーションの状態 */
typedef enum
{
    QRDSTATUS_IDLE,         /**< デコード待機 */
    QRDSTATUS_START,        /**< デコードスタート */
    QRDSTATUS_DECODE,       /**< デコード中 */
    QRDSTATUS_PREVIEW       /**< デコード結果表示、中断状態 */
} QRDStatus;

/** @brief カメライベントの状態 */
enum
{
    EVENT_CAMERA_TRANSFER = 0,
    EVENT_CAMERA_ERROR,
    EVENT_CAMERA_NUM
};

GraphicsControl g_GraphicsControl;  /** 描画操作クラス */

nn::fnd::ExpHeap s_AppHeap;         /** アプリが使用するヒープ */

using namespace mw::qrd;

// Y2R で RGB に変換したデータを受け取るバッファ
u8 *g_RgbBuffer;
u8 *g_TmpRgbBuffer;
// カメラのキャプチャデータを受信するバッファ
u8 *g_YuvBuffer;
u8 *g_TmpYuvBuffer;

int g_top, g_left, g_right, g_bottom;

/// サイズ
size_t g_RgbBytes;
size_t g_CameraOutputSize;
size_t g_TransferUnit;
u32 g_Y2rOutputSize;

// アプリ状態
static QRDStatus status_flag = QRDSTATUS_IDLE;

/**
 * @brief 画面に矩形のガイド線を表示する
 *
 */
void StampGuide()
{
    int rate = 50;
    int cutWidth = CAMERA_WIDTH / 2;
    int cut_x = (IMAGE_WIDTH - cutWidth) / 2;
    int cutHeight = CAMERA_HEIGHT / 2;
    int cut_y = (IMAGE_HEIGHT - cutHeight) / 2;

    // イメージの開始点
    int start_x = (nn::gx::DISPLAY0_HEIGHT - (IMAGE_WIDTH * rate / 100)) / 2;
    // ガイドの開始点
    int rect_x = start_x + (cut_x * rate / 100);
    // ガイドの幅
    int rectWidth = cutWidth * rate / 100;
    // イメージの開始点
    int start_y = (nn::gx::DISPLAY0_WIDTH - (IMAGE_HEIGHT * rate / 100)) / 2;
    // ガイドの開始点
    int rect_y = start_y + (cut_y * rate / 100);
    // ガイドの高さ
    int rectHeight = cutHeight * rate / 100;

    int top    = rect_y;
    int bottom = top + rectHeight;
    int left   = rect_x;
    int right  = left + rectWidth;

    int size = 20;
    f32 w = 2;
    f32 r = 1.0f;
    f32 g = 0.0f;
    f32 b = 0.0f;

    // 左上
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, left, top, left +size, top, w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, left, top, left, top +size, w, r, g, b);

    // 右上
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, right, top, right -10, top, w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, right, top, right, top +size, w, r, g, b);

    // 左下
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, left, bottom, left +size, bottom, w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, left, bottom, left, bottom -size, w, r, g, b);

    // 右上
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, right, bottom, right -size, bottom, w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, right, bottom, right, bottom -size, w, r, g, b);
}

/**
 * @brief 画面にQRの位置情報を表示する
 *
 * @param[in] top QRコード上辺位置
 * @param[in] left QRコード左辺位置
 * @param[in] right QRコード右辺位置
 * @param[in] bottom QRコード下辺位置
 */
void DrawLine(int top, int left, int right, int bottom)
{
    // イメージの開始点
    int rate = 50;
    int start_x = (nn::gx::DISPLAY0_HEIGHT - (CAMERA_WIDTH * rate / 100)) / 2;
    int startX = (left * rate / 100) + start_x;
    int startY = top  * rate / 100;
    int endX   = startX + (right - left) * rate / 100;
    int endY   = startY + (bottom - top) * rate / 100;
    f32 w = 2;
    f32 r = 0.0f;
    f32 g = 0.0f;
    f32 b = 1.0f;

    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, startX, startY, endX,   startY, w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, startX, startY, startX, endY  , w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, endX,   startY, endX,   endY  , w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, startX, endY,   endX,   endY  , w, r, g, b);
}

/**
 * @brief QRデコード実行
 *
 * @param[in] decoder QRデコードライブラリ
 * @return QRデコード成功時:true, 失敗時:falseを返す
 */
bool ExecuteDecode(QRDecoder *decoder)
{
    bool bRet = true;
    DecodeData res;
    ImageInfo info;
    int ret;

    memset(&res, 0, sizeof(res));
    memset(&info, 0, sizeof(info));

    info.width      = (unsigned short)CAMERA_WIDTH;
    info.height     = (unsigned short)CAMERA_HEIGHT;
    info.yuv422Data = (u32 *)g_YuvBuffer;

    // デコード位置を設定
    info.qrPosition.top    = (CAMERA_HEIGHT - IMAGE_HEIGHT) / 2;
    info.qrPosition.left   = (CAMERA_WIDTH  - IMAGE_WIDTH)  / 2;
    info.qrPosition.right  = info.qrPosition.left + IMAGE_WIDTH;
    info.qrPosition.bottom = info.qrPosition.top  + IMAGE_HEIGHT;

    ret = decoder->ExecuteDecode(&res, &info);

    if (!ret)
    {
        // QRデコード成功
        // QR バージョン
        NN_LOG("  QR Version: %d\n", res.version);
        // コンテンツ サイズ
        NN_LOG("  Contents size: %dbyte\n", res.size);
        // コンテンツ データ
        NN_LOG("  Contents data: %s\n", res.data);
        // QRコード位置
        NN_LOG("  qrPosition: top:%d, left:%d, right:%d, bottom:%d\n",
            res.qrPosition.top, res.qrPosition.left, res.qrPosition.right, res.qrPosition.bottom);
    	g_top = res.qrPosition.top;
    	g_left = res.qrPosition.left;
    	g_right = res.qrPosition.right;
    	g_bottom = res.qrPosition.bottom;
    }
    else
    {
        // QRデコード失敗
        status_flag = QRDSTATUS_START;
        bRet = false;
    }

    return bRet;
}

/**
 * @brief カメラの設定を行う
 *
 * @param[in] decoder QRデコードライブラリ
 * @return QRデコード成功時:true, 失敗時:falseを返す
 */
void CameraStart()
{
    memset(g_RgbBuffer, 0, CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_RGB_BYTES * sizeof(u8));
    memset(g_YuvBuffer, 0, CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_YUV_BYTES * sizeof(u8));

    // バッファにためるライン数を指定
    s16 transferLine = nn::camera::GetMaxLines(CAMERA_WIDTH, CAMERA_HEIGHT);
    nn::camera::SetTransferLines(nn::camera::PORT_CAM1, transferLine, CAMERA_WIDTH, CAMERA_HEIGHT);

    // フレームレートの変更
    nn::camera::SetFrameRate(nn::camera::SELECT_OUT1, nn::camera::FRAME_RATE_15); 

    // 画像サイズの変更
    nn::camera::SetSize(nn::camera::SELECT_OUT1, nn::camera::SIZE_VGA, nn::camera::CONTEXT_A);

    // カメラを起動状態にする
    nn::camera::Activate(nn::camera::SELECT_OUT1);
    nn::camera::SetTrimming(nn::camera::PORT_CAM1, false);

    // 変換中なら強制停止する。
    nn::y2r::StopConversion();
    while(nn::y2r::IsBusyConversion()) // Enableの間は設定できないので待つ
    {
    }
    // 入力フォーマットの変更
    nn::y2r::SetInputFormat(nn::y2r::INPUT_YUV422_BATCH);

    // 出力フォーマットの変更
    nn::y2r::SetOutputFormat(nn::y2r::OUTPUT_RGB_24);

    g_RgbBytes = nn::y2r::GetOutputFormatBytes(nn::y2r::OUTPUT_RGB_24);

    // 出力データの回転角度の変更
    nn::y2r::SetRotation(nn::y2r::ROTATION_NONE);

    // 出力データの並び順の変更
    nn::y2r::SetBlockAlignment(nn::y2r::BLOCK_LINE);

    // 入力データのサイズを指定
    nn::y2r::SetInputLineWidth(CAMERA_WIDTH);
    nn::y2r::SetInputLines(CAMERA_HEIGHT);

    // 変換係数の設定
    nn::y2r::SetStandardCoefficient(nn::y2r::COEFFICIENT_ITU_R_BT_601);

    // 転送する画像サイズからループ回数を計算
    g_TransferUnit = nn::camera::GetTransferBytes(nn::camera::PORT_CAM1);

    // 1フレームのバイト数
    g_CameraOutputSize = nn::camera::GetFrameBytes(CAMERA_WIDTH, CAMERA_HEIGHT);
    g_Y2rOutputSize = nn::y2r::GetOutputImageSize(CAMERA_WIDTH, CAMERA_HEIGHT,
                                                  nn::y2r::OUTPUT_RGB_24);
}

/**
 * @brief カメラのエラー/データ転送処理用スレッド
 *
 */
void CameraThreadFunc( uptr param )
{
    nn::os::Event  **events  = reinterpret_cast<nn::os::Event**>(param);

    // 最初の転送を設定する
    nn::camera::SetReceiving(events[EVENT_CAMERA_TRANSFER],
                             reinterpret_cast<void*>(g_TmpYuvBuffer),
                             nn::camera::PORT_CAM1, g_CameraOutputSize, g_TransferUnit);
    while(1)
    {
        s32 num = nn::os::WaitObject::WaitAny(reinterpret_cast<nn::os::WaitObject**>(events),
                                              EVENT_CAMERA_NUM);
        if (num == EVENT_CAMERA_TRANSFER)
        {
#if 0
            nn::camera::SetReceiving(events[EVENT_CAMERA_TRANSFER],
                                     reinterpret_cast<void*>(g_TmpYuvBuffer),
                                     nn::camera::PORT_CAM1, g_CameraOutputSize, g_TransferUnit);

            // 受信したYUVデータをYUVtoRGBで変換する。
            nn::y2r::SetSendingYuv(reinterpret_cast<void*>(g_TmpYuvBuffer),
                                   g_CameraOutputSize, CAMERA_WIDTH * Y2R_YUV_BYTES);
            nn::y2r::SetReceiving(reinterpret_cast<void*>(g_RgbBuffer),
                                  g_Y2rOutputSize, CAMERA_WIDTH * 8* g_RgbBytes);
            nn::y2r::StopConversion();
            while(nn::y2r::IsBusyConversion()) // Enableの間は設定できないので待つ
            {
            }
			nn::Result res = nn::y2r::StartConversion();
            if (res.IsSuccess()) {
				memcpy(g_TmpRgbBuffer, g_RgbBuffer, CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_RGB_BYTES);
			}

            if (QRDSTATUS_START == status_flag)
            {
                status_flag = QRDSTATUS_DECODE;
                memcpy(g_YuvBuffer, g_TmpYuvBuffer, g_CameraOutputSize);
            }
#else
            if (QRDSTATUS_START == status_flag) {
                status_flag = QRDSTATUS_DECODE;
            }
			memcpy(g_YuvBuffer, g_TmpYuvBuffer, g_CameraOutputSize);
			nn::camera::SetReceiving(events[EVENT_CAMERA_TRANSFER],
                                     reinterpret_cast<void*>(g_TmpYuvBuffer),
                                     nn::camera::PORT_CAM1, g_CameraOutputSize, g_TransferUnit);

            nn::y2r::SetSendingYuv(reinterpret_cast<void*>(g_YuvBuffer),
                                   g_CameraOutputSize, CAMERA_WIDTH * Y2R_YUV_BYTES);
            nn::y2r::SetReceiving(reinterpret_cast<void*>(g_RgbBuffer),
                                  g_Y2rOutputSize, CAMERA_WIDTH * 8* g_RgbBytes);
            nn::y2r::StopConversion();
            while(nn::y2r::IsBusyConversion()) // Enableの間は設定できないので待つ
            {
				nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMicroSeconds(100));
            }
			nn::Result res = nn::y2r::StartConversion();
            if (res.IsSuccess()) {
				memcpy(g_TmpRgbBuffer, g_RgbBuffer, CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_RGB_BYTES);
			}
#endif
        }
        else if (num == EVENT_CAMERA_ERROR)
        {
            // キャプチャ再開
            nn::camera::StopCapture(nn::camera::PORT_CAM1);
            nn::camera::ClearBuffer(nn::camera::PORT_CAM1);

            nn::camera::SetReceiving(events[EVENT_CAMERA_TRANSFER],
                                     reinterpret_cast<void*>(g_TmpYuvBuffer),
                                     nn::camera::PORT_CAM1, g_CameraOutputSize, g_TransferUnit);
            nn::camera::StartCapture(nn::camera::PORT_CAM1);
        }
    }
}

/**
 * @brief カメラのVSYNC処理用スレッド作成
 *
 */
void CameraVsyncThreadFunc( uptr param )
{
    nn::os::Event  *vsyncEvent  = reinterpret_cast<nn::os::Event*>(param);

    while(1)
    {
        vsyncEvent->Wait();
    }
}

/**
 * @brief メイン関数
 *
 */
void nnMain()
{
    bool bExit = false;     // 終了フラグ
    nn::Result result;      // 処理結果
    u32 mem_size = 0;       // QRデコード実行作業用領域バッファサイズ
    u32 *heap = NULL;       // QRデコード実行作業領域用バッファ

    QRDecoder *decoder = NULL;

    // os の初期化
    // nn::os::Initialize();

    // ヒープの確保
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(),
						 nn::os::GetDeviceMemorySize(),
						 nn::os::ALLOCATE_OPTION_LINEAR);

	// アプレットの初期化
	nn::applet::CTR::Enable();

    // グラフィックスの初期化
    g_GraphicsControl.InitializeGraphics();

	// 画像用のワークメモリを確保
	g_RgbBuffer = reinterpret_cast<u8 *>(s_AppHeap.Allocate(CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_RGB_BYTES));
    if (g_RgbBuffer == NULL)
    {
        NN_PANIC("ExpHeap.Allocate Error\n\n");
    }
	g_TmpRgbBuffer = reinterpret_cast<u8 *>(s_AppHeap.Allocate(CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_RGB_BYTES));
    if (g_TmpRgbBuffer == NULL)
    {
        NN_PANIC("ExpHeap.Allocate Error\n\n");
    }
	g_YuvBuffer = reinterpret_cast<u8 *>(s_AppHeap.Allocate(CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_YUV_BYTES));
    if (g_YuvBuffer == NULL)
    {
        NN_PANIC("ExpHeap.Allocate Error\n\n");
    }
	g_TmpYuvBuffer = reinterpret_cast<u8 *>(s_AppHeap.Allocate(CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_YUV_BYTES));
    if (g_TmpYuvBuffer == NULL)
    {
        NN_PANIC("ExpHeap.Allocate Error\n\n");
    }

    // CAMERA 初期化
    nn::camera::Initialize();

    // Y2R 初期化
    nn::y2r::Initialize();

    nn::os::Event errorIntr, vsyncIntr;

    // 割り込みハンドルの取得
    nn::camera::GetBufferErrorInterruptEvent(&errorIntr, nn::camera::PORT_CAM1);
    nn::camera::GetVsyncInterruptEvent(&vsyncIntr, nn::camera::PORT_CAM1);

    // カメラ関連設定
    CameraStart();

    nn::os::Event transEndIntr;
    // 割り込みハンドルの取得
    nn::y2r::GetTransferEndEvent(&transEndIntr);

    // hid の初期化
    result = nn::hid::Initialize();
    NN_UTIL_PANIC_IF_FAILED(result);

    // パッド準備
    nn::hid::PadReader padReader;
    nn::hid::PadStatus padStatus;

    // カメラのエラー/データ転送処理用スレッド作成
    nn::os::Event dataTransferIntr;

    nn::os::Thread cameraThread;
    nn::os::Event *events[EVENT_CAMERA_NUM] = {&dataTransferIntr, &errorIntr};
    cameraThread.StartUsingAutoStack(
        CameraThreadFunc,
        reinterpret_cast<uptr>(events),
        4096,
        nn::os::Thread::GetCurrentPriority()-2 );

    // カメラのVSYNC処理用スレッド作成
    nn::os::Thread cameraVsyncThread;
    cameraVsyncThread.StartUsingAutoStack(
        CameraVsyncThreadFunc,
        reinterpret_cast<uptr>(&vsyncIntr),
        4096,
        nn::os::Thread::GetCurrentPriority()-1 );

	// QRデコード実行作業用領域バッファサイズを取得
    mem_size = QRDecoder::GetDecodeBufferSize(CAMERA_WIDTH, CAMERA_HEIGHT);
    if (mem_size == 0)
    {
        NN_PANIC("failed QRDecoder::GetDecodeBufferSize().\n");
    }

    // QRデコード実行に必要な作業用バッファを用意
    heap = reinterpret_cast<u32*>(s_AppHeap.Allocate(mem_size));
    if (heap == NULL)
    {
        NN_PANIC("ExpHeap.Allocate Error\n\n");
    }

    // QRデコード初期化
    decoder = new QRDecoder;
    if (!decoder)
    {
        NN_PANIC("Could not initialize QRD library...\n");
    }
    decoder->Initialize(heap, mem_size);

    // キャプチャ開始
    nn::camera::ClearBuffer(nn::camera::PORT_CAM1);
    nn::camera::StartCapture(nn::camera::PORT_CAM1);

    NN_LOG("\nBUTTON_A : Start QRDecode\n");
	
	bool res_draw = false;

    while(!bExit)
    {
        // 入力取得
        padReader.ReadLatest(&padStatus);

        transEndIntr.Wait();
        {

            if (QRDSTATUS_IDLE == status_flag || QRDSTATUS_START == status_flag)
            {
                // 描画初期化
                g_GraphicsControl.ClearDisplay(NN_GX_DISPLAY0);
                // カメラ描画
                g_GraphicsControl.DrawCamera(g_TmpRgbBuffer, CAMERA_WIDTH, CAMERA_HEIGHT);
                // デコード位置の矩形指定をガイドで表示する
                StampGuide();
            }
        	
        	if (QRDSTATUS_PREVIEW == status_flag && res_draw) {
        		res_draw = false;
                // 描画初期化
                g_GraphicsControl.ClearDisplay(NN_GX_DISPLAY0);
                // カメラ描画
                g_GraphicsControl.DrawCamera(g_TmpRgbBuffer, CAMERA_WIDTH, CAMERA_HEIGHT);
                // デコード位置の矩形指定をガイドで表示する
                StampGuide();

	        	// イメージ位置描画
     		    DrawLine(g_top, g_left, g_right, g_bottom);
        	}

            switch (status_flag)
            {
            case QRDSTATUS_IDLE:
                if (padStatus.trigger & nn::hid::BUTTON_A)
                {
                    status_flag = QRDSTATUS_START;
                }
                break;
            case QRDSTATUS_START:
                // QRデコード開始時
                // カメラバッファからデコードするイメージデータを保持
                break;
            case QRDSTATUS_DECODE:
                // QRデコード実行
                if (ExecuteDecode(decoder))
                {
                    // デコードに成功
                    status_flag = QRDSTATUS_PREVIEW;
                	res_draw = true;
                }
                else
                {
                    // デコード失敗
                    status_flag = QRDSTATUS_START;
                }
                break;
            case QRDSTATUS_PREVIEW:
                if (padStatus.trigger & nn::hid::BUTTON_A)
                {
                    // QRデコード再開
                    status_flag = QRDSTATUS_IDLE;
                    NN_LOG("\nBUTTON_A : Start QRDecode\n");
                }
                break;
            }
            // 描画結果表示
            g_GraphicsControl.SwapBuffer(NN_GX_DISPLAY0, false);
            g_GraphicsControl.ClearDisplay(NN_GX_DISPLAY1);
            g_GraphicsControl.SwapBuffer();
        }
    }

    // グラフィックスの終了処理
    g_GraphicsControl.FinalizeGraphics();

    // カメラの終了処理
    StopCapture(nn::camera::PORT_CAM1);
    ClearBuffer(nn::camera::PORT_CAM1);
    nn::y2r::StopConversion();
    nn::y2r::Finalize();
    nn::camera::Finalize();

    // QRデコード終了
    decoder->Finalize();

    // ヒープの破棄
    if (NULL != heap)
    {
        s_AppHeap.Free(heap);
    }

    if (NULL != g_TmpYuvBuffer)
    {
        s_AppHeap.Free(g_TmpYuvBuffer);
    }
    if (NULL != g_YuvBuffer)
    {
        s_AppHeap.Free(g_YuvBuffer);
    }
    if (NULL != g_TmpRgbBuffer)
    {
        s_AppHeap.Free(g_TmpRgbBuffer);
    }
    if (NULL != g_RgbBuffer)
    {
        s_AppHeap.Free(g_RgbBuffer);
    }
    s_AppHeap.Finalize();
}
